@@ -3,7 +3,7 @@
 # Univention Python
 #  LDAP access
 #
-# Copyright 2002-2017 Univention GmbH
+# Copyright 2002-2019 Univention GmbH
 #
 # http://www.univention.de/
 #
@@ -31,28 +31,64 @@
 # <http://www.gnu.org/licenses/>.
 
 import re
+from functools import wraps
+
 import ldap
 import ldap.schema
 import ldap.sasl
-import univention.debug
-from univention.config_registry import ConfigRegistry
 from ldapurl import LDAPUrl
 from ldapurl import isLDAPUrl
 
+import univention.debug
+from univention.config_registry import ConfigRegistry
+try:
+	from typing import Any, Dict, List, Optional, Set, Tuple, Union  # noqa
+except ImportError:
+	pass
+
 
 def parentDn(dn, base=''):
-	_d = univention.debug.function('uldap.parentDn dn=%s base=%s' % (dn, base))
+	# type: (str, str) -> Optional[str]
+	"""
+	Return the parent container of a distinguished name.
+
+	:param str dn: The distinguished name.
+	:param str base: distinguished name where to stop.
+	:return: The parent distinguished name or None.
+	:rtype: str or None
+	"""
+	_d = univention.debug.function('uldap.parentDn dn=%s base=%s' % (dn, base))  # noqa F841
 	if dn.lower() == base.lower():
-		return
+		return None
 	dn = ldap.dn.str2dn(dn)
 	return ldap.dn.dn2str(dn[1:])
 
 
 def explodeDn(dn, notypes=0):
+	# type: (str, int) -> List[str]
+	"""
+	Break up a DN into its component parts.
+
+	:param str dn: The distinguished name.
+	:param int notypes: Return only the component's attribute values if True. Also the attribute types if False.
+	:return: A list of relative distinguished names.
+	:rtype: list[str]
+	"""
 	return ldap.dn.explode_dn(dn, notypes)
 
 
 def getRootDnConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
+	# type: (int, List[str], bool) -> access
+	"""
+	Open a LDAP connection to the local LDAP server with the LDAP root account.
+
+	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
+	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
+	:type decode_ignorelist: list[str]
+	:param bool reconnect: Automatically reconect if the connection fails.
+	:return: A LDAP access object.
+	:rtype: univention.uldap.access
+	"""
 	ucr = ConfigRegistry()
 	ucr.load()
 	port = int(ucr.get('slapd/port', '7389').split(',')[0])
@@ -61,12 +97,23 @@ def getRootDnConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 		bindpw = open('/etc/ldap.secret').read().rstrip('\n')
 		binddn = 'cn=admin,{0}'.format(ucr['ldap/base'])
 	else:
-		bindpw = open('/etc/ldap/rootpw.conf').read().rstrip('\n').lstrip('rootpw "').rstrip('"')
+		bindpw = open('/etc/ldap/rootpw.conf').read().rstrip('\n').replace('rootpw "', '', 1)[:-1]
 		binddn = 'cn=update,{0}'.format(ucr['ldap/base'])
 	return access(host=host, port=port, base=ucr['ldap/base'], binddn=binddn, bindpw=bindpw, start_tls=start_tls, decode_ignorelist=decode_ignorelist, reconnect=reconnect)
 
 
 def getAdminConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
+	# type: (int, List[str], bool) -> access
+	"""
+	Open a LDAP connection to the Master LDAP server using the admin credentials.
+
+	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
+	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
+	:type decode_ignorelist: list[str]
+	:param bool reconnect: Automatically reconect if the connection fails.
+	:return: A LDAP access object.
+	:rtype: univention.uldap.access
+	"""
 	ucr = ConfigRegistry()
 	ucr.load()
 	bindpw = open('/etc/ldap.secret').read().rstrip('\n')
@@ -75,6 +122,17 @@ def getAdminConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 
 
 def getBackupConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
+	# type: (int, List[str], bool) -> access
+	"""
+	Open a LDAP connection to a Backup LDAP server using the admin credentials.
+
+	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
+	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
+	:type decode_ignorelist: list[str]
+	:param bool reconnect: Automatically reconect if the connection fails.
+	:return: A LDAP access object.
+	:rtype: univention.uldap.access
+	"""
 	ucr = ConfigRegistry()
 	ucr.load()
 	bindpw = open('/etc/ldap-backup.secret').read().rstrip('\n')
@@ -89,6 +147,19 @@ def getBackupConnection(start_tls=2, decode_ignorelist=[], reconnect=True):
 
 
 def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, secret_file="/etc/machine.secret", reconnect=True):
+	# type: (int, List[str], bool, str, bool) -> access
+	"""
+	Open a LDAP connection using the machine credentials.
+
+	:param int start_tls: Negotiate TLS with server. If `2` is given, the command will require the operation to be successful.
+	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
+	:type decode_ignorelist: list[str]
+	:param bool ldap_master: Open a connection to the Master if True, to the preferred LDAP server otherwise.
+	:param str secret_file: The name of a file containing the password credentials.
+	:param bool reconnect: Automatically reconect if the connection fails.
+	:return: A LDAP access object.
+	:rtype: univention.uldap.access
+	"""
 	ucr = ConfigRegistry()
 	ucr.load()
 
@@ -116,10 +187,48 @@ def getMachineConnection(start_tls=2, decode_ignorelist=[], ldap_master=True, se
 			raise exc
 
 
-class access:
+def _fix_reconnect_handling(func):
+	# Bug #47926: python ldap does not reconnect on ldap.UNAVAILABLE
+	# We need this until https://github.com/python-ldap/python-ldap/pull/267 is fixed
+	@wraps(func)
+	def _decorated(self, *args, **kwargs):
+		if not self.reconnect:
+			return func(self, *args, **kwargs)
+
+		try:
+			return func(self, *args, **kwargs)
+		except ldap.INSUFFICIENT_ACCESS:
+			if self.whoami():  # the connection is still bound and valid
+				raise
+			self._reconnect()
+			return func(self, *args, **kwargs)
+		except (ldap.UNAVAILABLE, ldap.CONNECT_ERROR, ldap.TIMEOUT):  # ldap.TIMELIMIT_EXCEEDED ?
+			self._reconnect()
+			return func(self, *args, **kwargs)
+
+	return _decorated
+
+
+class access(object):
+	"""
+	The low-level class to access a LDAP server.
+
+	:param str host: host name of the LDAP server.
+	:param int port: TCP port of the LDAP server. Defaults to 7389 or 7636.
+	:param str base: LDAP base distinguished name.
+	:param str binddn: Distinguished name for simple authentication.
+	:param str bindpw: Password for simple authentication.
+	:param int start_tls: 0=no, 1=try StartTLS, 2=require StartTLS.
+	:param str ca_certfile: File name to CA certificate.
+	:param decode_ignorelist: List of LDAP attribute names which shall be handled as binary attributes.
+	:param use_ldaps bool: Connect to SSL port.
+	:param uri str: LDAP connection string.
+	:param follow_referral bool: Follow referrals and return result from other servers instead of returning the referral itself.
+	:param reconnect bool: Automatically re-establish connection to LDAP server if connection breaks.
+	"""
 
 	def __init__(self, host='localhost', port=None, base='', binddn='', bindpw='', start_tls=2, ca_certfile=None, decode_ignorelist=[], use_ldaps=False, uri=None, follow_referral=False, reconnect=True):
-		"""start_tls = 0 (no); 1 (try); 2 (must)"""
+		# type: (str, int, str, Optional[str], str, int, str, List, bool, str, bool, bool) -> None
 		self.host = host
 		self.base = base
 		self.binddn = binddn
@@ -171,15 +280,28 @@ class access:
 		else:
 			return pwd
 
+	@_fix_reconnect_handling
 	def bind(self, binddn, bindpw):
-		"""Do simple LDAP bind using DN and password."""
+		# type: (str, str) -> None
+		"""
+		Do simple LDAP bind using DN and password.
+
+		:param str binddn: The distinguished name of the account.
+		:param str bindpw: The user password for simple authentication.
+		"""
 		self.binddn = binddn
 		self.bindpw = bindpw
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'bind binddn=%s' % self.binddn)
 		self.lo.simple_bind_s(self.binddn, self.__encode_pwd(self.bindpw))
 
+	@_fix_reconnect_handling
 	def bind_saml(self, bindpw):
-		"""Do LDAP bind using SAML Message"""
+		# type: (str) -> None
+		"""
+		Do LDAP bind using SAML message.
+
+		:param str bindpw: The SAML authentication cookie.
+		"""
 		self.binddn = None
 		self.bindpw = bindpw
 		saml = ldap.sasl.sasl({
@@ -187,14 +309,35 @@ class access:
 			ldap.sasl.CB_PASS: bindpw,
 		}, 'SAML')
 		self.lo.sasl_interactive_bind_s('', saml)
-		self.binddn = re.sub('^dn:', '', self.lo.whoami_s())
+		self.binddn = self.whoami()
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'SAML bind binddn=%s' % self.binddn)
 
 	def unbind(self):
+		# type: () -> None
+		"""
+		Unauthenticate.
+		"""
 		self.lo.unbind_s()
 
+	def whoami(self):
+		# type: () -> str
+		"""
+		Return the distinguished name of the authenticated user.
+
+		:returns: The distinguished name.
+		:rtype: str
+		"""
+		dn = self.lo.whoami_s()
+		return re.sub('^dn:', '', dn)
+
+	def _reconnect(self):
+		# type: () -> None
+		"""Reconnect."""
+		self.lo.reconnect(self.lo._uri, retry_max=self.lo._retry_max, retry_delay=self.lo._retry_delay)
+
 	def __open(self, ca_certfile):
-		_d = univention.debug.function('uldap.__open host=%s port=%d base=%s' % (self.host, self.port, self.base))
+		# type: (Optional[str]) -> None
+		_d = univention.debug.function('uldap.__open host=%s port=%d base=%s' % (self.host, self.port, self.base))  # noqa F841
 
 		if self.reconnect:
 			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'establishing new connection with retry_max=%d' % self. client_connection_attempt)
@@ -209,15 +352,14 @@ class access:
 		if self.protocol.lower() != 'ldaps':
 			if self.start_tls == 1:
 				try:
-					self.lo.start_tls_s()
+					self.__starttls()
 				except:
 					univention.debug.debug(univention.debug.LDAP, univention.debug.WARN, 'Could not start TLS')
 			elif self.start_tls == 2:
-				self.lo.start_tls_s()
+				self.__starttls()
 
 		if self.binddn and not self.uri.startswith('ldapi://'):
-			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'bind binddn=%s' % self.binddn)
-			self.lo.simple_bind_s(self.binddn, self.__encode_pwd(self.bindpw))
+			self.bind(self.binddn, self.bindpw)
 
 		# Override referral handling
 		if self.follow_referral:
@@ -226,6 +368,10 @@ class access:
 		self.__schema = None
 		self.__reconnects_done = 0
 
+	@_fix_reconnect_handling
+	def __starttls(self):
+		self.lo.start_tls_s()
+
 	def __encode(self, value):
 		if value is None:
 			return value
@@ -265,9 +411,20 @@ class access:
 	def __decode_attribute(self, attr, val):
 		return self.__recode_attribute(attr, val)
 
+	@_fix_reconnect_handling
 	def get(self, dn, attr=[], required=False):
-		'''returns ldap object'''
+		# type: (str, List[str], bool) -> Dict[str, List[str]]
+		"""
+		Return multiple attributes of a single LDAP object.
 
+		:param str dn: The distinguished name of the object to lookup.
+		:param attr: The list of attributes to fetch.
+		:type attr: list[str]
+		:param bool required: Raise an exception instead of returning an empty dictionary.
+		:returns: A dictionary mapping the requested attributes to a list of their values.
+		:rtype: dict[str, list[str]]
+		:raises ldap.NO_SUCH_OBJECT: If the LDAP object is not accessible.
+		"""
 		if dn:
 			try:
 				result = self.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', attr)
@@ -279,10 +436,24 @@ class access:
 			raise ldap.NO_SUCH_OBJECT({'desc': 'no object'})
 		return {}
 
+	@_fix_reconnect_handling
 	def getAttr(self, dn, attr, required=False):
-		'''return attribute of ldap object'''
+		# type: (str, str, bool) -> List[str]
+		"""
+		Return a single attribute of a single LDAP object.
 
-		_d = univention.debug.function('uldap.getAttr %s %s' % (dn, attr))
+		:param str dn: The distinguished name of the object to lookup.
+		:param str attr: The attribute to fetch.
+		:param bool required: Raise an exception instead of returning an empty dictionary.
+		:returns: A list of values.
+		:rtype: list[str]
+		:raises ldap.NO_SUCH_OBJECT: If the LDAP object is not accessible.
+
+		.. warning:: the attribute name is currently case sensitive and must be given as in the LDAP schema
+
+		.. warning:: when `required=True` it raises `ldap.NO_SUCH_OBJECT` even if the object exists but the attribute is not set
+		"""
+		_d = univention.debug.function('uldap.getAttr %s %s' % (dn, attr))  # noqa F841
 		if dn:
 			try:
 				result = self.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', [attr])
@@ -294,10 +465,31 @@ class access:
 			raise ldap.NO_SUCH_OBJECT({'desc': 'no object'})
 		return []
 
-	def search(self, filter='(objectClass=*)', base='', scope='sub', attr=[], unique=False, required=False, timeout=-1, sizelimit=0, serverctrls=None):
-		'''do ldap search'''
+	@_fix_reconnect_handling
+	def search(self, filter='(objectClass=*)', base='', scope='sub', attr=[], unique=False, required=False, timeout=-1, sizelimit=0, serverctrls=None, response=None):
+		# type: (str, str, str, List[str], bool, bool, int, int, Optional[List[ldap.controls.LDAPControl]], Optional[Dict]) -> List[Tuple[str, Dict[str, List[str]]]]
+		"""
+		Perform LDAP search and return values.
 
-		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.search filter=%s base=%s scope=%s attr=%s unique=%d required=%d timeout=%d sizelimit=%d' % (filter, base, scope, attr, unique, required, timeout, sizelimit))
+		:param str filter: LDAP search filter.
+		:param str base: the starting point for the search.
+		:param str scope: Specify the scope of the search to be one of `base`, `base+one`, `one`, `sub`, or `domain` to specify a base object, base plus one-level, one-level, subtree, or children search.
+		:param attr: The list of attributes to fetch.
+		:type attr: list[str]
+		:param bool unique: Raise an exception if more than one object matches.
+		:param bool required: Raise an exception instead of returning an empty dictionary.
+		:param int timeout: wait at most timeout seconds for a search to complete. `-1` for no limit.
+		:param int sizelimit: retrieve at most sizelimit entries for a search. `0` for no limit.
+		:param serverctrls: a list of :py:class:`ldap.controls.LDAPControl` instances sent to the server along with the LDAP request.
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		:returns: A list of 2-tuples (dn, values) for each LDAP object, where values is a dictionary mapping attribute names to a list of values.
+		:rtype: list[tuple[str, dict[str, list[str]]]]
+		:raises ldap.NO_SUCH_OBJECT: Indicates the target object cannot be found.
+		:raises ldap.INAPPROPRIATE_MATCHING: Indicates that the matching rule specified in the search filter does not match a rule defined for the attribute's syntax.
+		"""
+		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.search filter=%s base=%s scope=%s attr=%s unique=%d required=%d timeout=%d sizelimit=%d' % (
+			filter, base, scope, attr, unique, required, timeout, sizelimit))
 
 		if not base:
 			base = self.base
@@ -320,16 +512,47 @@ class access:
 			raise ldap.NO_SUCH_OBJECT({'desc': 'no object'})
 		return res
 
-	def searchDn(self, filter='(objectClass=*)', base='', scope='sub', unique=False, required=False, timeout=-1, sizelimit=0, serverctrls=None):
-		_d = univention.debug.function('uldap.searchDn filter=%s base=%s scope=%s unique=%d required=%d' % (filter, base, scope, unique, required))
-		return [x[0] for x in self.search(filter, base, scope, ['dn'], unique, required, timeout, sizelimit, serverctrls)]
+	def searchDn(self, filter='(objectClass=*)', base='', scope='sub', unique=False, required=False, timeout=-1, sizelimit=0, serverctrls=None, response=None):
+		# type: (str, str, str, bool, bool, int, int, Optional[List[ldap.controls.LDAPControl]]) -> List[str]
+		"""
+		Perform LDAP search and return distinguished names only.
 
+		:param str filter: LDAP search filter.
+		:param str base: the starting point for the search.
+		:param str scope: Specify the scope of the search to be one of `base`, `base+one`, `one`, `sub`, or `domain` to specify a base object, base plus one-level, one-level, subtree, or children search.
+		:param bool unique: Raise an exception if more than one object matches.
+		:param bool required: Raise an exception instead of returning an empty dictionary.
+		:param int timeout: wait at most timeout seconds for a search to complete. `-1` for no limit.
+		:param int sizelimit: retrieve at most sizelimit entries for a search. `0` for no limit.
+		:param serverctrls: a list of :py:class:`ldap.controls.LDAPControl` instances sent to the server along with the LDAP request.
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		:returns: A list of distinguished names.
+		:rtype: list[str]
+		:raises ldap.NO_SUCH_OBJECT: Indicates the target object cannot be found.
+		:raises ldap.INAPPROPRIATE_MATCHING: Indicates that the matching rule specified in the search filter does not match a rule defined for the attribute's syntax.
+		"""
+		_d = univention.debug.function('uldap.searchDn filter=%s base=%s scope=%s unique=%d required=%d' % (filter, base, scope, unique, required))  # noqa F841
+		return [x[0] for x in self.search(filter, base, scope, ['dn'], unique, required, timeout, sizelimit, serverctrls, response)]
+
+	@_fix_reconnect_handling
 	def getPolicies(self, dn, policies=None, attrs=None, result=None, fixedattrs=None):
+		# type: (str, List[str], Dict[str, List[Any]], Any, Any) -> Dict[str, Dict[str, Any]]
+		"""
+		Return |UCS| policies for |LDAP| entry.
+
+		:param str dn: The distinguished name of the |LDAP| entry.
+		:param list policies: List of policy object classes...
+		:param dict attrs: |LDAP| attributes. If not given, the data is fetched from LDAP.
+		:param result: UNUSED!
+		:param fixedattrs: UNUSED!
+		:returns: A mapping of policy names to
+		"""
 		if attrs is None:
 			attrs = {}
 		if policies is None:
 			policies = []
-		_d = univention.debug.function('uldap.getPolicies dn=%s policies=%s attrs=%s' % (
+		_d = univention.debug.function('uldap.getPolicies dn=%s policies=%s attrs=%s' % (  # noqa F841
 			dn, policies, attrs))
 		if not dn and not policies:  # if policies is set apply a fictionally referenced list of policies
 			return {}
@@ -367,6 +590,15 @@ class access:
 		return result
 
 	def _merge_policy(self, policy_dn, obj_dn, object_classes, result):
+		# type: (str, str, Set[str], Dict[str, Dict[str, Any]]) -> None
+		"""
+		Merge policies into result.
+
+		:param policy_dn str: Distinguished name of the policy object.
+		:param obj_dn: Distinguished name of the LDAP object.
+		:param object_classes set: the set of object classes of the LDAP object.
+		:param result list: A mapping, into which the policy is merged.
+		"""
 		pattrs = self.get(policy_dn)
 		if not pattrs:
 			return
@@ -406,7 +638,15 @@ class access:
 					'fixed': 1 if key in fixed else 0,
 				}
 
+	@_fix_reconnect_handling
 	def get_schema(self):
+		# type: () -> ldap.schema.subentry.SubSchema
+		"""
+		Retrieve |LDAP| schema information from |LDAP| server.
+
+		:returns: The |LDAP| schema.
+		:rtype: ldap.schema.subentry.SubSchema
+		"""
 		if self.reconnect and self.lo._reconnects_done > self.__reconnects_done:
 			# the schema might differ after reconnecting (e.g. slapd restart)
 			self.__schema = None
@@ -415,13 +655,23 @@ class access:
 			self.__schema = ldap.schema.SubSchema(self.lo.read_subschemasubentry_s(self.lo.search_subschemasubentry_s()), 0)
 		return self.__schema
 
+	@_fix_reconnect_handling
 	def add(self, dn, al, serverctrls=None, response=None):
-		"""Add LDAP entry with dn and attributes in add_list=(attribute-name, old-values. new-values) or (attribute-name, new-values)."""
+		# type: (str, List[Tuple], Optional[List[ldap.controls.LDAPControl]], Optional[dict]) -> None
+		"""
+		Add LDAP entry at distinguished name and attributes in add_list=(attribute-name, old-values. new-values) or (attribute-name, new-values).
+
+		:param str dn: The distinguished name of the object to add.
+		:param al: The add-list of 2-tuples (attribute-name, new-values).
+		:param serverctrls: a list of ldap.controls.LDAPControl instances sent to the server along with the LDAP request
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		"""
 		if not serverctrls:
 			serverctrls = []
 
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.add dn=%s' % dn)
-		nal = {}
+		nal = {}  # type: Dict[str, Any]
 		for i in al:
 			key, val = i[0], i[-1]
 			if not val:
@@ -444,9 +694,20 @@ class access:
 		if serverctrls and isinstance(response, dict):
 			response['ctrls'] = resp_ctrls
 
+	@_fix_reconnect_handling
 	def modify(self, dn, changes, serverctrls=None, response=None):
-		"""Modify LDAP entry dn with attributes in changes=(attribute-name, old-values, new-values)."""
+		# type: (str, List[Tuple[str, Any, Any]], Optional[List[ldap.controls.LDAPControl]], Optional[dict]) -> str
+		"""
+		Modify LDAP entry DN with attributes in changes=(attribute-name, old-values, new-values).
 
+		:param str dn: The distinguished name of the object to modify.
+		:param changes: The modify-list of 3-tuples (attribute-name, old-values, new-values).
+		:param serverctrls: a list of ldap.controls.LDAPControl instances sent to the server along with the LDAP request
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		:returns: The distinguished name.
+		:rtype: str
+		"""
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.modify %s' % dn)
 
 		if not serverctrls:
@@ -508,8 +769,15 @@ class access:
 			return ldap.dn.dn2str([ldap.dn.str2dn(new_rdn)[0]] + ldap.dn.str2dn(dn)[1:]), new_rdn
 		return dn, rdn
 
+	@_fix_reconnect_handling
 	def modify_s(self, dn, ml):
-		"""Redirect modify_s directly to lo"""
+		# type: (str, List[Tuple[str, Optional[List[str]], List[str]]]) -> None
+		"""
+		Redirect `modify_s` directly to :py:attr:`lo`.
+
+		:param str dn: The distinguished name of the object to modify.
+		:param ml: The modify-list of 3-tuples (attribute-name, old-values, new-values).
+		"""
 		try:
 			self.lo.modify_ext_s(dn, ml)
 		except ldap.REFERRAL as exc:
@@ -518,8 +786,18 @@ class access:
 			lo_ref = self._handle_referral(exc)
 			lo_ref.modify_ext_s(dn, ml)
 
+	@_fix_reconnect_handling
 	def modify_ext_s(self, dn, ml, serverctrls=None, response=None):
-		"""Redirect modify_ext_s directly to lo"""
+		# type: (str, List[Tuple[str, Any, Any]], Optional[List[ldap.controls.LDAPControl]], Optional[dict]) -> None
+		"""
+		Redirect `modify_ext_s` directly to :py:attr:`lo`.
+
+		:param str dn: The distinguished name of the object to modify.
+		:param ml: The modify-list of 3-tuples (attribute-name, old-values, new-values).
+		:param serverctrls: a list of ldap.controls.LDAPControl instances sent to the server along with the LDAP request
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		"""
 		if not serverctrls:
 			serverctrls = []
 
@@ -535,6 +813,16 @@ class access:
 			response['ctrls'] = resp_ctrls
 
 	def rename(self, dn, newdn, serverctrls=None, response=None):
+		# type: (str, str, Optional[List[ldap.controls.LDAPControl]], Optional[dict]) -> None
+		"""
+		Rename a LDAP object.
+
+		:param str dn: The old distinguished name of the object to rename.
+		:param str newdn: The new distinguished name of the object to rename.
+		:param serverctrls: a list of ldap.controls.LDAPControl instances sent to the server along with the LDAP request
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		"""
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.rename %s -> %s' % (dn, newdn))
 		oldsdn = self.parentDn(dn)
 		newrdn = ldap.dn.dn2str([ldap.dn.str2dn(newdn)[0]])
@@ -550,8 +838,19 @@ class access:
 			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.rename: modrdn %s to %s' % (dn, newrdn))
 			self.rename_ext_s(dn, newrdn, serverctrls=serverctrls, response=response)
 
+	@_fix_reconnect_handling
 	def rename_ext_s(self, dn, newrdn, newsuperior=None, serverctrls=None, response=None):
-		"""Redirect rename_ext_s directly to lo"""
+		# type: (str, str, Optional[str], Optional[List[ldap.controls.LDAPControl]], Optional[dict]) -> None
+		"""
+		Redirect `rename_ext_s` directly to :py:attr:`lo`.
+
+		:param str dn: The old distinguished name of the object to rename.
+		:param str newdn: The new distinguished name of the object to rename.
+		:param str newsuperior: The distinguished name of the new container.
+		:param serverctrls: a list of ldap.controls.LDAPControl instances sent to the server along with the LDAP request
+		:type serverctrls: list[ldap.controls.LDAPControl]
+		:param dict response: An optional dictionary to receive the server controls of the result.
+		"""
 		if not serverctrls:
 			serverctrls = []
 
@@ -566,7 +865,14 @@ class access:
 		if serverctrls and isinstance(response, dict):
 			response['ctrls'] = resp_ctrls
 
+	@_fix_reconnect_handling
 	def delete(self, dn):
+		# type: (str) -> None
+		"""
+		Delete a LDAP object.
+
+		:param str dn: The distinguished name of the object to remove.
+		"""
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'uldap.delete %s' % dn)
 		if dn:
 			univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'delete')
@@ -579,15 +885,38 @@ class access:
 				lo_ref.delete_s(dn)
 
 	def parentDn(self, dn):
+		# type: (str) -> Optional[str]
+		"""
+		Return the parent container of a distinguished name.
+
+		:param str dn: The distinguished name.
+		:return: The parent distinguished name or None if the LDAP base is reached.
+		:rtype: str or None
+		"""
 		return parentDn(dn, self.base)
 
 	def explodeDn(self, dn, notypes=False):
+		# type: (str, Union[bool, int]) -> List[str]
+		"""
+		Break up a DN into its component parts.
+
+		:param str dn: The distinguished name.
+		:param bool notypes: Return only the component's attribute values if True. Also the attribute types if False.
+		:return: A list of relative distinguished names.
+		:rtype: list[str]
+		"""
 		return explodeDn(dn, notypes)
 
 	@classmethod
 	def compare_dn(cls, a, b):
+		# type: (str, str) -> bool
 		r"""Test DNs are same
 
+		:param str a: The first distinguished name.
+		:param str b: A second distinguished name.
+		:returns: True if the DNs are the same, False otherwise.
+		:rtype: bool
+
 		>>> compare_dn = access.compare_dn
 		>>> compare_dn('foo=1', 'foo=1')
 		True
@@ -613,17 +942,31 @@ class access:
 		return [sorted((x.lower(), y, z) for x, y, z in rdn) for rdn in ldap.dn.str2dn(a)] == [sorted((x.lower(), y, z) for x, y, z in rdn) for rdn in ldap.dn.str2dn(b)]
 
 	def __getstate__(self):
-		_d = univention.debug.function('uldap.__getstate__')
+		"""
+		Return state for pickling.
+		"""
+		_d = univention.debug.function('uldap.__getstate__')  # noqa F841
 		odict = self.__dict__.copy()
 		del odict['lo']
 		return odict
 
 	def __setstate__(self, dict):
-		_d = univention.debug.function('uldap.__setstate__')
+		"""
+		Set state for pickling.
+		"""
+		_d = univention.debug.function('uldap.__setstate__')  # noqa F841
 		self.__dict__.update(dict)
 		self.__open(self.ca_certfile)
 
 	def _handle_referral(self, exception):
+		# type: (ldap.REFERRAL) -> ldap.ldapobject.ReconnectLDAPObject
+		"""
+		Follow LDAP rederral.
+
+		:param exception ldap.REFERRAL: The LDAP referral exception.
+		:returns: LDAP connection object for the referred LDAP server.
+		:rtype: ldap.ldapobject.ReconnectLDAPObject
+		"""
 		univention.debug.debug(univention.debug.LDAP, univention.debug.INFO, 'Following LDAP referral')
 		exc = exception.args[0]
 		info = exc.get('info')
@@ -631,6 +974,7 @@ class access:
 		if isLDAPUrl(ldap_url):
 			conn_str = LDAPUrl(ldap_url).initializeUrl()
 
+			# FIXME?: this upgrades a access(reconnect=False) connection to a reconnect=True connection
 			lo_ref = ldap.ldapobject.ReconnectLDAPObject(conn_str, trace_stack_limit=None)
 
 			if self.ca_certfile:
